/********************************************************************\
 * This program is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU General Public License as   *
 * published by the Free Software Foundation; either version 2 of   *
 * the License, or (at your option) any later version.              *
 *                                                                  *
 * This program is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public License*
 * along with this program; if not, contact:                        *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
 *                                                                  *
\********************************************************************/

/* $Id: conf.h 1162 2007-01-06 23:51:02Z benoitg $ */
/** @file conf.h
    @brief Config file parsing
    @author Copyright (C) 2004 Philippe April <papril777@yahoo.com>
*/

#ifndef _CONFIG_H_
#define _CONFIG_H_

/*@{*/ 
/** Defines */
/** How long till we give up detecting the interface with the default route */
#define NUM_EXT_INTERFACE_DETECT_RETRY 120
/** How often should we try to detect the interface with the default route
 *  if it isn't up yet */
#define EXT_INTERFACE_DETECT_RETRY_INTERVAL 1

/** Defaults configuration values */
#ifndef SYSCONFDIR
	#define DEFAULT_CONFIGFILE "/etc/nodogsplash/nodogsplash.conf"
#else
	#define DEFAULT_CONFIGFILE SYSCONFDIR"/nodogsplash/nodogsplash.conf"
#endif	
#define DEFAULT_DAEMON 1
#define DEFAULT_DEBUGLEVEL LOG_INFO
#define DEFAULT_MAXCLIENTS 20
#define DEFAULT_GATEWAYNAME "NoDogSplash"
#define DEFAULT_GATEWAYPORT 2050
#define DEFAULT_CHECKINTERVAL 60
#define DEFAULT_CLIENTTIMEOUT 10
#define DEFAULT_CLIENTFORCEOUT 360
#define DEFAULT_WEBROOT "/etc/nodogsplash/htdocs"
#define DEFAULT_SPLASHPAGE "splash.html"
#define DEFAULT_IMAGESDIR "images"
#define DEFAULT_AUTHDIR "nodogsplash_auth"
#define DEFAULT_DENYDIR "nodogsplash_deny"
#define DEFAULT_TRAFFIC_CONTROL 0
#define DEFAULT_UPLOAD_LIMIT 0
#define DEFAULT_DOWNLOAD_LIMIT 0
#define DEFAULT_LOG_SYSLOG 0
#define DEFAULT_SYSLOG_FACILITY LOG_DAEMON
#define DEFAULT_NDSCTL_SOCK "/tmp/ndsctl.sock"
#define DEFAULT_INTERNAL_SOCK "/tmp/ndsctl.sock"
/*@}*/ 

/**
 * Firewall rules
 */
typedef struct _firewall_rule_t {
    int block_allow;		/**< @brief 1 = Allow rule, 0 = Block rule */
    char *protocol;		/**< @brief tcp, udp, etc ... */
    char *port;			/**< @brief Port to block/allow */
    char *mask;			/**< @brief Mask for the rule *destination* */
    struct _firewall_rule_t *next;
} t_firewall_rule;

/**
 * Firewall rulesets
 */
typedef struct _firewall_ruleset_t {
    char			*name;
    t_firewall_rule		*rules;
    struct _firewall_ruleset_t	*next;
} t_firewall_ruleset;

/**
 * MAC Addresses
 */
typedef struct _MAC_t {
    char   *mac;
    struct _MAC_t *next;
} t_MAC;


/**
 * Configuration structure
 */
typedef struct {
  char configfile[255];	/**< @brief name of the config file */
  char *ndsctl_sock;		/**< @brief ndsctl path to socket */
  char *internal_sock;		/**< @brief internal path to socket */
  int daemon;			/**< @brief if daemon > 0, use daemon mode */
  int debuglevel;		/**< @brief Debug information verbosity */
  int maxclients;		/**< @brief Maximum number of clients allowed */
  char *ext_interface;		/**< @brief Interface to external network */
  char *gw_name;		/**< @brief Name of the gateway; e.g. its SSID */
  char *gw_interface;		/**< @brief Interface we will accept connections on */
  char *gw_address;		/**< @brief Internal IP address for our web server */
  int gw_port;			/**< @brief Port the webserver will run on */
  char *webroot;		/**< @brief Directory containing splash pages, etc. */
  char *splashpage;		/**< @brief Name of main splash page */
  char *imagesdir;		/**< @brief Subdir of webroot containing .png .gif files etc */
  char *redirectURL;		/**< @brief URL to direct client to after authentication */
  char *authdir;		/**< @brief Notional relative dir for authentication URL */
  char *denydir;		/**< @brief Notional relative dir for denial URL */
  int clienttimeout;		/**< @brief How many CheckIntervals before an inactive client
				   must be re-authenticated */
  int clientforceout;		/**< @brief How many CheckIntervals before a client
				   must be re-authenticated */
  int checkinterval;		/**< @brief Period the the client timeout check
				   thread will run, in seconds */
  int traffic_control;		/**< @brief boolean, whether to do tc */
  int download_limit;		/**< @brief Download limit, kb/s */
  int upload_limit;		/**< @brief Upload limit, kb/s */
  int log_syslog;		/**< @brief boolean, whether to log to syslog */
  int syslog_facility;		/**< @brief facility to use when using syslog for
			   	logging */
  t_firewall_ruleset *rulesets;	/**< @brief firewall rules */
  t_MAC *trustedmaclist; 	/**< @brief list of trusted macs */
  t_MAC *blockedmaclist; 	/**< @brief list of blocked macs */
} s_config;

/** @brief Get the current gateway configuration */
s_config *config_get_config(void);

/** @brief Initialise the conf system */
void config_init(void);

/** @brief Initialize the variables we override with the command line*/
void config_init_override(void);

/** @brief Reads the configuration file */
void config_read(char *filename);

/** @brief Check that the configuration is valid */
void config_validate(void);

/** @brief Fetch a firewall rule set. */
t_firewall_rule *get_ruleset(char *);

static void config_notnull(void *parm, char *parmname);
static int parse_boolean_value(char *);
static int _parse_firewall_rule(char *ruleset, char *leftover);
static void parse_firewall_ruleset(char *, FILE *, char *, int *);
void parse_trusted_mac_list(char *);
void parse_blocked_mac_list(char *);

#define LOCK_CONFIG() do { \
	debug(LOG_DEBUG, "Locking config"); \
	pthread_mutex_lock(&config_mutex); \
	debug(LOG_DEBUG, "Config locked"); \
} while (0)

#define UNLOCK_CONFIG() do { \
	debug(LOG_DEBUG, "Unlocking config"); \
	pthread_mutex_unlock(&config_mutex); \
	debug(LOG_DEBUG, "Config unlocked"); \
} while (0)

#endif /* _CONFIG_H_ */
